/****************************************************************************
 *                                                                          *
 * Filename:    mdnspd.c                                                    *
 * Author:      Tabitha Tipper <t.tipper@lancaster.ac.uk>                      *
 * Created:     21/11/03                                                    *
 * Modified:    01/01/04                                                    *
 * Description: The Mangling DNS Proxy Daemon listens on localhost:53 and   *
 *              passes on all requests to real DNS, if it sees a record     *
 *              returned with an AAAA record it maps the v6 address to an   *
 *              v4 address (see dnsrh) and creates an A record to return.   *
 *                                                                          *
 * (c) 2003 Tabitha Tipper, licensed under the GNU General Public License      *
 *     See the LICENSE file for details.                                    *
 ****************************************************************************/

/* FIXME: Add signal catching to re-read configs and die cleanly etc */
/* FIXME: Remove hard-coded NS addresses, read from file */

/* Signal catching not yet implemented 
#include <errno.h>
#include <signal.h>
#include <unistd.h> */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>

/* Networking headers */
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
/* #include <sys/wait.h> */

#include <semaphore.h>

#include "mdnspd-support.h"

int main(int argc, char * argv[]) { 

  /* struct semaphore realSem; */
  sem_t * realSem;
  int srvSock = 0;
  struct sockaddr_in sai;
  int err = 0;
  char upstreamNS[100];

  /* setup semaphore */
  realSem = (sem_t *) malloc(sizeof(sem_t));
  sem_init(realSem, 
      0, 
      (unsigned int) 1);

  /* setup upstream NS */
  memset(upstreamNS, '\0', 100);
  if (argc > 1) { 
    strncpy(upstreamNS, argv[1], 100);
  } else { 
    strncpy(upstreamNS, DEFAULT_UPSTREAM, 100);
  }

  /* Setup the connection to LHOST:PORTNUM for the mdnspd */
  srvSock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (-1 == srvSock)
    fprintf(stderr, "mdnspd: Error creating socket.\n");

  sai.sin_family = AF_INET;
  sai.sin_port = htons(PORTNUM);
  sai.sin_addr.s_addr = INADDR_ANY;
  memset(&sai.sin_zero, 0, sizeof(sai.sin_zero));

  err = bind(srvSock, (struct sockaddr *) &sai, sizeof(struct sockaddr_in));
  if (-1 == err) 
    fprintf(stderr, "mdnspd: Error binding to port.\n");

  if (DEBUG) 
    fprintf(stderr, "mdnspd: Socket created, looping for input.\n");

  /* The main worker loop, accepts connections and spawns threads to send them
   * off too, then loops.
   *
   * Cannot be exited cleanly via signal catching at the minute.
   */
  while (1) { 
    pthread_t conn = 0;		    /* Thread itself */
    int connRet = 0;		    /* return of thread */
    struct sockaddr_in clientAddr;  /* info about client */
    connectData * cd = NULL;	    /* struct of info to pass to each thread */
    int addrlen = 0;		    /* length of a sockaddr_in struct */
    int reqLen = 0;		    /* length of the incoming query */
    unsigned char request[512];	    /* buffer for incoming query */


    addrlen = sizeof(struct sockaddr_in);
    memset(request, 0, 512);

    /* read max 512 from the client and pass it up to the server, 
     * try A first then AAAA type... but only if A requested first off.
     */
    reqLen = recvfrom(srvSock, request, 512, 0, 
	(struct sockaddr *) &clientAddr, &addrlen);

    /* Construct a new connectData struct and pass it off to a new thread */
    cd = (connectData *) malloc(sizeof(connectData));

    cd->sem = realSem;

    cd->serverSocket = &srvSock;

    /* cd->request = (char *) malloc(reqLen); */
    cd->request = (char *) malloc(512);
    memset(cd->request, '0', 512);
    memcpy(cd->request, request, reqLen);

    cd->reqLen = reqLen;

    cd->upstreamNS = (char *) malloc(sizeof(upstreamNS));
    memcpy(cd->upstreamNS, upstreamNS, sizeof(upstreamNS));

/*    memcpy(&cd->clientAddr, &clientAddr, addrlen); */
    cd->clientAddr.sin_family = clientAddr.sin_family;
    cd->clientAddr.sin_port = clientAddr.sin_port;
    cd->clientAddr.sin_addr = clientAddr.sin_addr;

    fprintf(stderr, "mdnspd: request from '%s:%d'\n", inet_ntoa(cd->clientAddr.sin_addr), cd->clientAddr.sin_port);

    connRet = pthread_create(&conn, NULL, (void *)&connection, (void *) cd);
  }
    
  /* eventually catch a signal, break and quit */

  return 0;
}
